__title__: str = "pathy"
__version__: str = "0.10.3"
__summary__: str = "pathlib.Path subclasses for local and cloud bucket storage"
__uri__: str = "https://github.com/justindujardin/pathy"
__author__: str = "Justin DuJardin"
__email__: str = "justin@dujardinconsulting.com"
__license__: str = "Apache 2.0"
